/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/* graphical inlets and outlets, both for control and signals.  */

/* This code is highly inefficient; messages actually have to be forwarded
by inlets and outlets.  The outlet is in even worse shape than the inlet;
in order to avoid having a "signal" method in the class, the oulet actually
sprouts an inlet, which forwards the message to the "outlet" object, which
sends it on to the outlet proper.  Another way to do it would be to have
separate classes for "signal" and "control" outlets, but this would complicate
life elsewhere. */

/* changes by Thomas Musil IEM KUG Graz Austria 2001 */

/* improvement: inlets and outlets go graphical */
/* and are editable by properties */

/* all changes are labeled with      iemlib      */

#include "m_pd.h"
#include "g_canvas.h"
#include <string.h>
/*iemlib*/
#include <stdio.h>
#include "g_7_guis.h"

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif

#ifndef  PD_VERSION_MINOR
#define PD_VERSION_MINOR 32
#endif
/*iemlib*/

void signal_setborrowed(t_signal *sig, t_signal *sig2);
void signal_makereusable(t_signal *sig);


/* ------------------------- vinlet -------------------------- */
/*iemlib*/
t_widgetbehavior vinlet_widgetbehavior;
/*iemlib*/
t_class *vinlet_class;

typedef struct _vinlet
{
    t_object x_obj;
    t_canvas *x_canvas;
    t_inlet *x_inlet;
    int x_bufsize;
    t_float *x_buf; 	    /* signal buffer; zero if not a signal */
    t_float *x_endbuf;
    t_float *x_fill;
    t_float *x_read;
    int x_hop;
    	/* if not reblocking, the next slot communicates the parent's inlet
	signal from the prolog to the DSP routine: */
    t_signal *x_directsignal;
    /*iemlib*/
    t_glist   *x_glist;
    int       x_size;
    int       x_ldx;
    int       x_ldy;
    char      x_font[16];
    int       x_fontsize;
    int       x_bcol;
    int       x_lcol;
    int       x_selected;
    t_symbol  *x_lab;
    /*iemlib*/
} t_vinlet;

static void *vinlet_new(t_symbol *s/*iemlib*/, int argc, t_atom *argv/*iemlib*/)
{
    t_vinlet *x = (t_vinlet *)pd_new(vinlet_class);
    /*iemlib*/
    int bcol= -262144;
    int lcol= -1;
    int a=IEM_GUI_DEFAULTSIZE;
    int ldx=8, ldy=-8;
    int f=0, fs=10;
    t_symbol *lab = gensym("empty");

    if((argc == 8)&&IS_A_FLOAT(argv,0)&&IS_A_SYMBOL(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,4)
       &&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)&&IS_A_FLOAT(argv,7))
    {
        a = (int)atom_getintarg(0, argc, argv);
	lab = atom_getsymbolarg(1, argc, argv);
	ldx = (int)atom_getintarg(2, argc, argv);
	ldy = (int)atom_getintarg(3, argc, argv);
        f = (int)atom_getintarg(4, argc, argv);
        fs = (int)atom_getintarg(5, argc, argv);
        bcol = (int)atom_getintarg(6, argc, argv);
        lcol = (int)atom_getintarg(7, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a&0xfffffffe);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    /*iemlib*/
    x->x_canvas = canvas_getcurrent();
    x->x_inlet = canvas_addinlet(x->x_canvas, &x->x_obj.ob_pd, 0);
    x->x_bufsize = 0;
    x->x_buf = 0;
    outlet_new(&x->x_obj, 0);
    return (x);
}

/*iemlib*/
/* widget helper functions */

void vinlet_draw(t_vinlet *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(firsttime)
    {
	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size,
		 x->x_bcol, x);
        sys_vgui(".x%x.c create line %d %d %d %d %d %d -tags %xV\n",
		 canvas, x->x_obj.te_xpos+1, x->x_obj.te_ypos+1,
		 x->x_obj.te_xpos + x->x_size/2, x->x_obj.te_ypos + x->x_size/2,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos, x);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c create text %d %d -text {~} -anchor w \
		 -font {%s %d bold} -tags %xTIL\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos+x->x_size/2+2,
		 x->x_font, x->x_size, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xOUT%d\n",
		 canvas, x->x_obj.te_xpos,
		 x->x_obj.te_ypos + x->x_size-1, x->x_obj.te_xpos + IOWIDTH,
		 x->x_obj.te_ypos + x->x_size, x, 0);
    }
    else
    {
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size);
        sys_vgui(".x%x.c coords %xV %d %d %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos+1, x->x_obj.te_ypos+1,
		 x->x_obj.te_xpos + x->x_size/2, x->x_obj.te_ypos + x->x_size/2,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c coords %xTIL %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos+x->x_size/2+2);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
        sys_vgui(".x%x.c coords %xOUT%d %d %d %d %d\n",
		 canvas, x, 0, x->x_obj.te_xpos,
		 x->x_obj.te_ypos + x->x_size-1, x->x_obj.te_xpos + IOWIDTH,
		 x->x_obj.te_ypos + x->x_size);
    }
}

void vinlet_erase(t_vinlet* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xV\n", canvas, x);
    if(x->x_buf != 0)
	sys_vgui(".x%x.c delete %xTIL\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xOUT%d\n", canvas, x, 0);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void vinlet_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_vinlet *x = (t_vinlet *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void vinlet_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_vinlet *x = (t_vinlet *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    vinlet_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void vinlet_select(t_gobj *z, t_glist *glist, int selected)
{
    t_vinlet *x = (t_vinlet *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
	sys_vgui(".x%x.c itemconfigure %xV -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c itemconfigure %xTIL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
	sys_vgui(".x%x.c itemconfigure %xV -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c itemconfigure %xTIL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void vinlet_delete(t_gobj *z, t_glist *glist)
{
    t_vinlet *x = (t_vinlet *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void vinlet_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_vinlet *x = (t_vinlet *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 vinlet_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 vinlet_erase(x, glist);
         rtext_free(y);
    }
}

static void vinlet_save(t_gobj *z, t_binbuf *b)
{
    t_vinlet *x = (t_vinlet *)z;
    int bcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisisiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		(x->x_buf != 0)?gensym("il~"):gensym("il"), x->x_size,
		x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8, x->x_fontsize,
		bcol, lcol);
    binbuf_addv(b, ";");
}

static void vinlet_properties(t_gobj *z, t_glist *owner)
{
    t_vinlet *x = (t_vinlet *)z;
    char buf[600];

    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s %s \
	    ----------dimensions(pix):----------- %g %g size: 0 0 empty \
	    empty 0.0 empty 0.0 empty %g \
	    %g empty empty %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
            (x->x_buf != 0)?"SIGNAL-INLET":"CONTROL-INLET",
	    (float)x->x_size, (float)IEM_GUI_MINSIZE,
            0.0,/*no_schedule*/
	    -1.0, -1.0, -1.0,/*no linlog, no init, no multi*/
	    "nosndno", "norcvno",/*no send, no receive*/
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, -1/*no front-color*/, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_lab);
}

static void vinlet_dialog(t_vinlet *x, t_symbol *s, int argc, t_atom *argv)
{
    if((argc == 17)&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,4)
       &&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)&&IS_A_SYMBOL(argv,7)
       &&IS_A_SYMBOL(argv,8)&&IS_A_SYMBOL(argv,9)&&IS_A_FLOAT(argv,10)
       &&IS_A_FLOAT(argv,11)&&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13)
       &&IS_A_FLOAT(argv,14)&&IS_A_FLOAT(argv,15)&&IS_A_FLOAT(argv,16))
    {
	t_canvas *canvas=glist_getcanvas(x->x_glist);
	int a = (int)atom_getintarg(0, argc, argv);
	t_symbol *lab = atom_getsymbolarg(9, argc, argv);
	int ldx = (int)atom_getintarg(10, argc, argv);
	int ldy = (int)atom_getintarg(11, argc, argv);
	int f = (int)atom_getintarg(12, argc, argv);
	int fs = (int)atom_getintarg(13, argc, argv);
	int bcol = (int)atom_getintarg(14, argc, argv);
	int lcol = (int)atom_getintarg(16, argc, argv);

	str_xch_raute2dollar(lab);
	x->x_size = iem_gui_clip_size(a&0xfffffffe);
	x->x_lcol = lcol & 0xffffff;
	x->x_bcol = bcol & 0xffffff;
	x->x_lab = lab;
	x->x_ldx = ldx;
	x->x_ldy = ldy;
	if(f == 1) strcpy(x->x_font, "helvetica");
	else if(f == 2) strcpy(x->x_font, "times");
	else strcpy(x->x_font, "courier");
	if(fs < 4)
	    fs = 4;
	x->x_fontsize = fs;

	if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		     x->x_lab->s_name);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c itemconfigure %xTIL -font {%s %d bold}\n",
		     canvas, x, x->x_font, x->x_size);
	sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x,
		 x->x_bcol);
	vinlet_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(canvas, (t_text*)x);
    }
    else
        outlet_anything(x->x_obj.ob_outlet, s, argc, argv);
}
/*iemlib*/

static void vinlet_bang(t_vinlet *x)
{
    outlet_bang(x->x_obj.ob_outlet);
}

static void vinlet_pointer(t_vinlet *x, t_gpointer *gp)
{
    outlet_pointer(x->x_obj.ob_outlet, gp);
}

static void vinlet_float(t_vinlet *x, t_float f)
{
    outlet_float(x->x_obj.ob_outlet, f);
}

static void vinlet_symbol(t_vinlet *x, t_symbol *s)
{
    outlet_symbol(x->x_obj.ob_outlet, s);
}

static void vinlet_list(t_vinlet *x, t_symbol *s, int argc, t_atom *argv)
{
    outlet_list(x->x_obj.ob_outlet, s, argc, argv);
}

static void vinlet_anything(t_vinlet *x, t_symbol *s, int argc, t_atom *argv)
{
    outlet_anything(x->x_obj.ob_outlet, s, argc, argv);
}

static void vinlet_free(t_vinlet *x)
{
    canvas_rminlet(x->x_canvas, x->x_inlet);
}

t_inlet *vinlet_getit(t_pd *x)
{
    if (pd_class(x) != vinlet_class) bug("vinlet_getit");
    return (((t_vinlet *)x)->x_inlet);
}

/* ------------------------- signal inlet -------------------------- */
int vinlet_issignal(t_vinlet *x)
{
    return (x->x_buf != 0);
}

static int tot;

t_int *vinlet_perform(t_int *w)
{
    t_vinlet *x = (t_vinlet *)(w[1]);
    t_float *out = (t_float *)(w[2]);
    int n = (int)(w[3]);
    t_float *in = x->x_read;
#if 0
    if (tot < 5) post("-in %x out %x n %d", in, out, n);
    if (tot < 5) post("-buf %x endbuf %x", x->x_buf, x->x_endbuf);
    if (tot < 5) post("in[0] %f in[1] %f in[2] %f", in[0], in[1], in[2]);
#endif
    while (n--) *out++ = *in++;
    if (in == x->x_endbuf) in = x->x_buf;
    x->x_read = in;
    return (w+4);
}

static void vinlet_dsp(t_vinlet *x, t_signal **sp)
{
    t_signal *outsig;
    	/* no buffer means we're not a signal inlet */
    if (!x->x_buf)
    	return;
    outsig = sp[0];
    if (x->x_directsignal)
    {
    	signal_setborrowed(sp[0], x->x_directsignal);
    }
    else
    {
	dsp_add(vinlet_perform, 3, x, outsig->s_vec, outsig->s_n);
	x->x_read = x->x_buf;
    }
}

    /* prolog code: loads buffer from parent patch */
t_int *vinlet_doprolog(t_int *w)
{
    t_vinlet *x = (t_vinlet *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    int n = (int)(w[3]);
    t_float *out = x->x_fill;
    if (out == x->x_endbuf)
    {
    	t_float *f1 = x->x_buf, *f2 = x->x_buf + x->x_hop;
    	int nshift = x->x_bufsize - x->x_hop;
    	out -= x->x_hop;
    	while (nshift--) *f1++ = *f2++;
    }
#if 0
    if (tot < 5) post("in %x out %x n %x", in, out, n), tot++;
    if (tot < 5) post("in[0] %f in[1] %f in[2] %f", in[0], in[1], in[2]);
#endif
    while (n--) *out++ = *in++;
    x->x_fill = out;
    return (w+4);
}

int inlet_getsignalindex(t_inlet *x);

    	/* set up prolog DSP code  */
void vinlet_dspprolog(t_vinlet *x, t_signal **parentsigs,
    int myvecsize, int phase, int period, int frequency, int reblock,
    	int switched)
{
    t_signal *insig, *outsig;

    	/* if the "reblock" flag is set, arrange to copy data in from the
	parent. */
    if (reblock)
    {
	int parentvecsize, bufsize, oldbufsize, prologphase;
	    /* this should never happen: */
	if (!x->x_buf) return;

    	    /* the prolog code counts from 0 to period-1; the
    	    phase is backed up by one so that AFTER the prolog code
    	    runs, the "x_fill" phase is in sync with the "x_read" phase. */

	prologphase = (phase - 1) & (period - 1);
	if (parentsigs)
	{
    	    insig = parentsigs[inlet_getsignalindex(x->x_inlet)];
    	    parentvecsize = insig->s_n;
	}
	else
	{
    	    insig = 0;
    	    parentvecsize = 1;
	}
	bufsize = parentvecsize;
	if (bufsize < myvecsize) bufsize = myvecsize;
	if (bufsize != (oldbufsize = x->x_bufsize))
	{
    	    t_float *buf = x->x_buf;
    	    t_freebytes(buf, oldbufsize * sizeof(*buf));
    	    buf = (t_float *)t_getbytes(bufsize * sizeof(*buf));
    	    memset((char *)buf, 0, bufsize * sizeof(*buf));
    	    x->x_bufsize = bufsize;
    	    x->x_endbuf = buf + bufsize;
    	    x->x_buf = buf;
	}
	if (parentsigs)
	{
    	    x->x_hop = period * parentvecsize;
    	    x->x_fill = x->x_endbuf -
    		(x->x_hop - prologphase * parentvecsize);
    	    dsp_add(vinlet_doprolog, 3, x, insig->s_vec, parentvecsize);
	    	/* if the input signal's reference count is zero, we have
	    	to free it here because we didn't in ugen_doit(). */
	    if (!insig->s_refcount)
	    	signal_makereusable(insig);
	}
	else memset((char *)(x->x_buf), 0, bufsize * sizeof(*x->x_buf));
	x->x_directsignal = 0;
    }
    else
    {
    	    /* no reblocking; in this case our output signal is "borrowed"
	    and merely needs to be pointed to the real one. */
    	x->x_directsignal = parentsigs[inlet_getsignalindex(x->x_inlet)];
    }
}

static void *vinlet_newsig(/*iemlib*/t_symbol *s, int argc, t_atom *argv/*iemlib*/)
{
    t_vinlet *x = (t_vinlet *)pd_new(vinlet_class);
    /*iemlib*/
    int bcol= -262144;
    int lcol= -1;
    int a=IEM_GUI_DEFAULTSIZE;
    int ldx=8, ldy=-8;
    int f=0, fs=10;
    t_symbol *lab = gensym("empty");

    if((argc == 8)&&IS_A_FLOAT(argv,0)&&IS_A_SYMBOL(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,4)
       &&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)&&IS_A_FLOAT(argv,7))
    {
        a = (int)atom_getintarg(0, argc, argv);
	lab = atom_getsymbolarg(1, argc, argv);
	ldx = (int)atom_getintarg(2, argc, argv);
	ldy = (int)atom_getintarg(3, argc, argv);
        f = (int)atom_getintarg(4, argc, argv);
        fs = (int)atom_getintarg(5, argc, argv);
        bcol = (int)atom_getintarg(6, argc, argv);
        lcol = (int)atom_getintarg(7, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
	fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a&0xfffffffe);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    /*iemlib*/
    x->x_canvas = canvas_getcurrent();
    x->x_inlet = canvas_addinlet(x->x_canvas, &x->x_obj.ob_pd, &s_signal);
    x->x_endbuf = x->x_buf = (t_float *)getbytes(0);
    x->x_bufsize = 0;
    x->x_directsignal = 0;
    outlet_new(&x->x_obj, &s_signal);
    return (x);
}

static void vinlet_setup(void)
{
    vinlet_class = class_new(gensym("inlet"), (t_newmethod)vinlet_new,
    	(t_method)vinlet_free, sizeof(t_vinlet), CLASS_NOINLET, /*iemlib*/A_GIMME/*iemlib*/, 0);
    class_addcreator((t_newmethod)vinlet_newsig, gensym("inlet~"), 0);
    /*iemlib*/
    class_addcreator((t_newmethod)vinlet_new, gensym("il"), A_GIMME, 0);
    class_addcreator((t_newmethod)vinlet_newsig, gensym("il~"), A_GIMME, 0);
    /*iemlib*/
    class_addbang(vinlet_class, vinlet_bang);
    class_addpointer(vinlet_class, vinlet_pointer);
    class_addfloat(vinlet_class, vinlet_float);
    class_addsymbol(vinlet_class, vinlet_symbol);
    class_addlist(vinlet_class, vinlet_list);
    class_addanything(vinlet_class, vinlet_anything);
    /*iemlib*/
    class_addmethod(vinlet_class, (t_method)vinlet_dialog, gensym("dialog"), A_GIMME, 0);
    /*iemlib*/
    class_addmethod(vinlet_class, (t_method)vinlet_dsp, gensym("dsp"), 0);
    /*iemlib*/
    vinlet_widgetbehavior.w_getrectfn = vinlet_getrect;
    vinlet_widgetbehavior.w_displacefn = vinlet_displace;
    vinlet_widgetbehavior.w_selectfn = vinlet_select;
    vinlet_widgetbehavior.w_activatefn = NULL;
    vinlet_widgetbehavior.w_deletefn = vinlet_delete;
    vinlet_widgetbehavior.w_visfn = vinlet_vis;
#if (PD_VERSION_MINOR > 31)
    vinlet_widgetbehavior.w_clickfn = NULL;
    vinlet_widgetbehavior.w_propertiesfn = vinlet_properties;
#endif
    vinlet_widgetbehavior.w_savefn = vinlet_save;
    class_setwidget(vinlet_class, &vinlet_widgetbehavior);
    /*iemlib*/
    class_sethelpsymbol(vinlet_class, gensym("pd"));
}

/* ------------------------- voutlet -------------------------- */
/*iemlib*/
t_widgetbehavior voutlet_widgetbehavior;
/*iemlib*/
t_class *voutlet_class;

typedef struct _voutlet
{
    t_object x_obj;
    t_canvas *x_canvas;
    t_outlet *x_parentoutlet;
    int x_bufsize;
    t_float *x_buf; 	    /* signal buffer; zero if not a signal */
    t_float *x_endbuf;
    t_float *x_empty;	    /* next to read out of buffer in epilog code */
    t_float *x_write;	    /* next to write in to buffer */
    int x_hop;	    	    /* hopsize */
    	/* vice versa from the inlet, if we don't block, this holds the
	parent's outlet signal, valid between the prolog and the dsp setup
	routines.  */
    t_signal *x_directsignal;
    	/* and here's a flag indicating that we aren't blocked but have to
	do a copy (because we're switched). */
    char x_justcopyout;
    /*iemlib*/
    t_glist   *x_glist;
    int       x_size;
    int       x_ldx;
    int       x_ldy;
    char      x_font[16];
    int       x_fontsize;
    int       x_bcol;
    int       x_lcol;
    int       x_selected;
    t_symbol  *x_lab;
    /*iemlib*/
} t_voutlet;

static void *voutlet_new(t_symbol *s/*iemlib*/, int argc, t_atom *argv/*iemlib*/)
{
    t_voutlet *x = (t_voutlet *)pd_new(voutlet_class);
    /*iemlib*/
    int bcol= -262144;
    int lcol= -1;
    int a=IEM_GUI_DEFAULTSIZE;
    int ldx=8, ldy=-8;
    int f=0, fs=10;
    t_symbol *lab = gensym("empty");

    if((argc == 8)&&IS_A_FLOAT(argv,0)&&IS_A_SYMBOL(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,4)
       &&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)&&IS_A_FLOAT(argv,7))
    {
        a = (int)atom_getintarg(0, argc, argv);
	lab = atom_getsymbolarg(1, argc, argv);
	ldx = (int)atom_getintarg(2, argc, argv);
	ldy = (int)atom_getintarg(3, argc, argv);
        f = (int)atom_getintarg(4, argc, argv);
        fs = (int)atom_getintarg(5, argc, argv);
        bcol = (int)atom_getintarg(6, argc, argv);
        lcol = (int)atom_getintarg(7, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
        fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a&0xfffffffe);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    /*iemlib*/
    x->x_canvas = canvas_getcurrent();
    x->x_parentoutlet = canvas_addoutlet(x->x_canvas, &x->x_obj.ob_pd, 0);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, 0, 0);
    x->x_bufsize = 0;
    x->x_buf = 0;
    return (x);
}

/*iemlib*/
/* widget helper functions */

void voutlet_draw(t_voutlet *x, t_glist *glist, int firsttime)
{
    t_canvas *canvas=glist_getcanvas(glist);

    if(firsttime)
    {
	sys_vgui(".x%x.c create rectangle %d %d %d %d -fill #%6.6x -tags %xBASE\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size,
		 x->x_bcol, x);
        sys_vgui(".x%x.c create line %d %d %d %d %d %d -tags %xA\n",
		 canvas, x->x_obj.te_xpos+1, x->x_obj.te_ypos+x->x_size-1,
		 x->x_obj.te_xpos + x->x_size/2, x->x_obj.te_ypos + x->x_size/2,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos+x->x_size, x);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c create text %d %d -text {~} -anchor w \
		     -font {%s %d bold} -tags %xTIL\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos+x->x_size/2-2,
		 x->x_font, x->x_size, x);
        if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     "", x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        else
            sys_vgui(".x%x.c create text %d %d -text {%s} -anchor w \
		     -font {%s %d bold} -fill #%6.6x -tags %xLABEL\n",
		     canvas, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy,
		     x->x_lab->s_name, x->x_font, x->x_fontsize,
		     x->x_lcol, x);
        sys_vgui(".x%x.c create rectangle %d %d %d %d -tags %xIN%d\n",
		 canvas, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + IOWIDTH, x->x_obj.te_ypos+1, x, 0);
    }
    else
    {
	sys_vgui(".x%x.c coords %xBASE %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos + x->x_size);
        sys_vgui(".x%x.c coords %xA %d %d %d %d %d %d\n",
		 canvas, x, x->x_obj.te_xpos+1, x->x_obj.te_ypos+x->x_size-1,
		 x->x_obj.te_xpos + x->x_size/2, x->x_obj.te_ypos + x->x_size/2,
		 x->x_obj.te_xpos + x->x_size, x->x_obj.te_ypos+x->x_size);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c coords %xTIL %d %d\n",
		 canvas, x, x->x_obj.te_xpos, x->x_obj.te_ypos+x->x_size/2-2);
        sys_vgui(".x%x.c coords %xLABEL %d %d\n",
		 canvas, x, x->x_obj.te_xpos+x->x_ldx, x->x_obj.te_ypos+x->x_ldy);
        sys_vgui(".x%x.c coords %xIN%d %d %d %d %d\n",
		 canvas, x, 0, x->x_obj.te_xpos, x->x_obj.te_ypos,
		 x->x_obj.te_xpos + IOWIDTH, x->x_obj.te_ypos+1);
    }
}

void voutlet_erase(t_voutlet* x, t_glist* glist)
{
    t_canvas *canvas=glist_getcanvas(glist);

    sys_vgui(".x%x.c delete %xBASE\n", canvas, x);
    sys_vgui(".x%x.c delete %xA\n", canvas, x);
    if(x->x_buf != 0)
	sys_vgui(".x%x.c delete %xTIL\n", canvas, x);
    sys_vgui(".x%x.c delete %xLABEL\n", canvas, x);
    sys_vgui(".x%x.c delete %xIN%d\n", canvas, x, 0);
}

/* ------------------------ bg widgetbehaviour----------------------------- */

static void voutlet_getrect(t_gobj *z, t_glist *owner, int *xp1, int *yp1, int *xp2, int *yp2)
{
    t_voutlet *x = (t_voutlet *)z;

    *xp1 = x->x_obj.te_xpos;
    *yp1 = x->x_obj.te_ypos;
    *xp2 = x->x_obj.te_xpos + x->x_size;
    *yp2 = x->x_obj.te_ypos + x->x_size;
}

static void voutlet_displace(t_gobj *z, t_glist *glist, int dx, int dy)
{
    t_voutlet *x = (t_voutlet *)z;

    x->x_obj.te_xpos += dx;
    x->x_obj.te_ypos += dy;
    voutlet_draw(x, glist, 0);
    canvas_fixlinesfor(glist_getcanvas(glist), (t_text *)x);
}

static void voutlet_select(t_gobj *z, t_glist *glist, int selected)
{
    t_voutlet *x = (t_voutlet *)z;
    t_canvas *canvas=glist_getcanvas(glist);

    x->x_selected = selected;
    if(selected)
    {
        pd_bind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xA -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c itemconfigure %xTIL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_SELECTED);
    }
    else
    {
        pd_unbind(&x->x_obj.ob_pd, iem_gui_key_sym);
	sys_vgui(".x%x.c itemconfigure %xBASE -outline #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xA -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c itemconfigure %xTIL -fill #%6.6x\n", canvas, x, IEM_GUI_COLOR_NORMAL);
        sys_vgui(".x%x.c itemconfigure %xLABEL -fill #%6.6x\n", canvas, x, x->x_lcol);
    }
}

static void voutlet_delete(t_gobj *z, t_glist *glist)
{
    t_voutlet *x = (t_voutlet *)z;

    canvas_deletelinesfor(glist_getcanvas(glist), (t_text *)z);
}

static void voutlet_vis(t_gobj *z, t_glist *glist, int vis)
{
    t_voutlet *x = (t_voutlet *)z;

    if(vis)
    {
         t_rtext *y = rtext_new_without_senditup(glist, (t_text *)z, glist->gl_editor->e_rtext);
	 voutlet_draw(x, glist, 1);
    }
    else
    {
         t_rtext *y = glist_findrtext(glist, (t_text *)z);
	 voutlet_erase(x, glist);
         rtext_free(y);
    }
}

static void voutlet_save(t_gobj *z, t_binbuf *b)
{
    t_voutlet *x = (t_voutlet *)z;
    int bcol, lcol;

    lcol = -1 - (((0xfc0000 & x->x_lcol) >> 6)|((0xfc00 & x->x_lcol) >> 4)|((0xfc & x->x_lcol) >> 2));
    bcol = -1 - (((0xfc0000 & x->x_bcol) >> 6)|((0xfc00 & x->x_bcol) >> 4)|((0xfc & x->x_bcol) >> 2));

    binbuf_addv(b, "ssiisisiiiiii", gensym("#X"),gensym("obj"),
		(t_int)x->x_obj.te_xpos, (t_int)x->x_obj.te_ypos,
		(x->x_buf != 0)?gensym("ol~"):gensym("ol"), x->x_size,
		x->x_lab,
		x->x_ldx, x->x_ldy,
                (int)(x->x_font[0]-'a'+1)/8, x->x_fontsize,
		bcol, lcol);
    binbuf_addv(b, ";");
}

static void voutlet_properties(t_gobj *z, t_glist *owner)
{
    t_voutlet *x = (t_voutlet *)z;
    char buf[600];

    str_xch_dollar2raute(x->x_lab);
    sprintf(buf, "pdtk_iemgui_dialog %%s %s \
	    ----------dimensions(pix):----------- %g %g size: 0 0 empty \
	    empty 0.0 empty 0.0 empty %g \
	    %g empty empty %g %g \
	    %s %s \
	    %s %g %g \
            %g %g \
	    %d %d %d\n",
            (x->x_buf != 0)?"SIGNAL-OUTLET":"CONTROL-OUTLET",
	    (float)x->x_size, (float)IEM_GUI_MINSIZE,
            0.0,/*no_schedule*/
	    -1.0, -1.0, -1.0,/*no linlog, no init, no multi*/
	    "nosndno", "norcvno",/*no send, no receive*/
	    x->x_lab->s_name, (float)x->x_ldx, (float)x->x_ldy,
            (float)((int)(x->x_font[0]-'a'+1)/8), (float)x->x_fontsize,
            0xffffff & x->x_bcol, -1/*no front-color*/, 0xffffff & x->x_lcol);
    gfxstub_new(&x->x_obj.ob_pd, x, buf);
    str_xch_raute2dollar(x->x_lab);
}

static void voutlet_dialog(t_voutlet *x, t_symbol *s, int argc, t_atom *argv)
{
    if((argc == 17)&&IS_A_FLOAT(argv,0)&&IS_A_FLOAT(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,4)
       &&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)&&IS_A_SYMBOL(argv,7)
       &&IS_A_SYMBOL(argv,8)&&IS_A_SYMBOL(argv,9)&&IS_A_FLOAT(argv,10)
       &&IS_A_FLOAT(argv,11)&&IS_A_FLOAT(argv,12)&&IS_A_FLOAT(argv,13)
       &&IS_A_FLOAT(argv,14)&&IS_A_FLOAT(argv,15)&&IS_A_FLOAT(argv,16))
    {
	t_canvas *canvas=glist_getcanvas(x->x_glist);
	int a = (int)atom_getintarg(0, argc, argv);
	t_symbol *lab = atom_getsymbolarg(9, argc, argv);
	int ldx = (int)atom_getintarg(10, argc, argv);
	int ldy = (int)atom_getintarg(11, argc, argv);
	int f = (int)atom_getintarg(12, argc, argv);
	int fs = (int)atom_getintarg(13, argc, argv);
	int bcol = (int)atom_getintarg(14, argc, argv);
	int lcol = (int)atom_getintarg(16, argc, argv);

	str_xch_raute2dollar(lab);
	x->x_size = iem_gui_clip_size(a&0xfffffffe);
	x->x_lcol = lcol & 0xffffff;
	x->x_bcol = bcol & 0xffffff;
	x->x_lab = lab;
	x->x_ldx = ldx;
	x->x_ldy = ldy;
	if(f == 1) strcpy(x->x_font, "helvetica");
	else if(f == 2) strcpy(x->x_font, "times");
	else strcpy(x->x_font, "courier");
	if(fs < 4)
	    fs = 4;
	x->x_fontsize = fs;

	if(!strcmp(x->x_lab->s_name, "empty"))
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol, "");
	else
	    sys_vgui(".x%x.c itemconfigure %xLABEL -font {%s %d bold} -fill #%6.6x -text {%s} \n",
		     canvas, x, x->x_font, x->x_fontsize,
		     x->x_selected?IEM_GUI_COLOR_SELECTED:x->x_lcol,
		     x->x_lab->s_name);
	if(x->x_buf != 0)
	    sys_vgui(".x%x.c itemconfigure %xTIL -font {%s %d bold}\n",
		     canvas, x, x->x_font, x->x_size);
	sys_vgui(".x%x.c itemconfigure %xBASE -fill #%6.6x\n", canvas, x,
		 x->x_bcol);
	voutlet_draw(x, x->x_glist, 0);
	canvas_fixlinesfor(canvas, (t_text*)x);
    }
    else
	outlet_anything(x->x_parentoutlet, s, argc, argv);
}
/*iemlib*/

static void voutlet_bang(t_voutlet *x)
{
    outlet_bang(x->x_parentoutlet);
}

static void voutlet_pointer(t_voutlet *x, t_gpointer *gp)
{
    outlet_pointer(x->x_parentoutlet, gp);
}

static void voutlet_float(t_voutlet *x, t_float f)
{
    outlet_float(x->x_parentoutlet, f);
}

static void voutlet_symbol(t_voutlet *x, t_symbol *s)
{
    outlet_symbol(x->x_parentoutlet, s);
}

static void voutlet_list(t_voutlet *x, t_symbol *s, int argc, t_atom *argv)
{
    outlet_list(x->x_parentoutlet, s, argc, argv);
}

static void voutlet_anything(t_voutlet *x, t_symbol *s, int argc, t_atom *argv)
{
    outlet_anything(x->x_parentoutlet, s, argc, argv);
}

static void voutlet_free(t_voutlet *x)
{
    canvas_rmoutlet(x->x_canvas, x->x_parentoutlet);
}

t_outlet *voutlet_getit(t_pd *x)
{
    if (pd_class(x) != voutlet_class) bug("voutlet_getit");
    return (((t_voutlet *)x)->x_parentoutlet);
}

/* ------------------------- signal outlet -------------------------- */

int voutlet_issignal(t_voutlet *x)
{
    return (x->x_buf != 0);
}

    /* LATER optimize for non-overlapped case where the "+=" isn't needed */
t_int *voutlet_perform(t_int *w)
{
    t_voutlet *x = (t_voutlet *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    int n = (int)(w[3]);
    t_float *out = x->x_write, *outwas = out;
#if 0
    if (tot < 5) post("-in %x out %x n %d", in, out, n);
    if (tot < 5) post("-buf %x endbuf %x", x->x_buf, x->x_endbuf);
#endif
    while (n--)
    {
    	*out++ += *in++;
    	if (out == x->x_endbuf) out = x->x_buf;
    }
    outwas += x->x_hop;
    if (outwas >= x->x_endbuf) outwas = x->x_buf;
    x->x_write = outwas;
    return (w+4);
}

    /* epilog code for blocking: write buffer to parent patch */
static t_int *voutlet_doepilog(t_int *w)
{
    t_voutlet *x = (t_voutlet *)(w[1]);
    t_float *out = (t_float *)(w[2]);
    int n = (int)(w[3]);
    t_float *in = x->x_empty;
#if 0
    if (tot < 5) post("outlet in %x out %x n %x", in, out, n), tot++;
#endif
    for (; n--; in++) *out++ = *in, *in = 0;
    if (in == x->x_endbuf) in = x->x_buf;
    x->x_empty = in;
    return (w+4);
}

int outlet_getsignalindex(t_outlet *x);

    	/* prolog for outlets -- store pointer to the outlet on the
	parent, which, if "reblock" is false, will want to refer
	back to whatever we see on our input during the "dsp" method
	called later.  */
void voutlet_dspprolog(t_voutlet *x, t_signal **parentsigs,
    int myvecsize, int phase, int period, int frequency, int reblock,
    	int switched)
{
    x->x_justcopyout = (switched && !reblock);
    if (reblock)
    {
    	x->x_directsignal = 0;
    }
    else
    {
    	if (!parentsigs) bug("voutlet_dspprolog");
    	x->x_directsignal =
	    parentsigs[outlet_getsignalindex(x->x_parentoutlet)];
    }
}

static void voutlet_dsp(t_voutlet *x, t_signal **sp)
{
    t_signal *insig;
    if (!x->x_buf) return;
    insig = sp[0];
    if (x->x_justcopyout)
    	dsp_add_copy(insig->s_vec, x->x_directsignal->s_vec, insig->s_n);
    else if (x->x_directsignal)
    {
    	    /* if we're just going to make the signal available on the
	    parent patch, hand it off to the parent signal. */
	/* this is done elsewhere--> sp[0]->s_refcount++; */
    	signal_setborrowed(x->x_directsignal, sp[0]);
    }
    else
    	dsp_add(voutlet_perform, 3, x, insig->s_vec, insig->s_n);
}

    	/* set up epilog DSP code.  If we're reblocking, this is the
	time to copy the samples out to the containing object's outlets.
	If we aren't reblocking, there's nothing to do here.  */
void voutlet_dspepilog(t_voutlet *x, t_signal **parentsigs,
    int myvecsize, int phase, int period, int frequency, int reblock,
    	int switched)
{
    if (!x->x_buf) return;  /* this shouldn't be necesssary... */
    if (reblock)
    {
	t_signal *insig, *outsig;
	int parentvecsize, bufsize, oldbufsize;
	int bigperiod, epilogphase, blockphase;
	if (parentsigs)
	{
    	    outsig = parentsigs[outlet_getsignalindex(x->x_parentoutlet)];
    	    parentvecsize = outsig->s_n;
	}
	else
	{
    	    outsig = 0;
    	    parentvecsize = 1;
	}
	bigperiod = myvecsize/parentvecsize;
	if (!bigperiod) bigperiod = 1;
	epilogphase = phase & (bigperiod - 1);
	blockphase = (phase + period - 1) & (bigperiod - 1) & (- period);
	bufsize = parentvecsize;
	if (bufsize < myvecsize) bufsize = myvecsize;
	if (bufsize != (oldbufsize = x->x_bufsize))
	{
    	    t_float *buf = x->x_buf;
    	    t_freebytes(buf, oldbufsize * sizeof(*buf));
    	    buf = (t_float *)t_getbytes(bufsize * sizeof(*buf));
    	    memset((char *)buf, 0, bufsize * sizeof(*buf));
    	    x->x_bufsize = bufsize;
    	    x->x_endbuf = buf + bufsize;
    	    x->x_buf = buf;
	}
	if (parentvecsize * period > bufsize) bug("voutlet_dspepilog");
	x->x_write = x->x_buf + parentvecsize * blockphase;
	if (period == 1 && frequency > 1)
    	    x->x_hop = parentvecsize/frequency;
    	else x->x_hop = period * parentvecsize;
	/* post("phase %d, block %d, parent %d", phase & 63,
    	    parentvecsize * blockphase, parentvecsize * epilogphase); */

	if (parentsigs)
	{
    		/* set epilog pointer and schedule it */
    	    x->x_empty = x->x_buf + parentvecsize * epilogphase;
    	    dsp_add(voutlet_doepilog, 3, x, outsig->s_vec, parentvecsize);
	}
    }
    	/* if we aren't blocked but we are switched, the epilog code just
	copies zeros to the output.  In this case the blocking code actually
	jumps over the epilog if the block is running. */
    else if (switched)
    {
    	if (parentsigs)
	{
    	    t_signal *outsig =
	    	parentsigs[outlet_getsignalindex(x->x_parentoutlet)];
	    dsp_add_zero(outsig->s_vec, outsig->s_n);
    	}
    }
}

static void *voutlet_newsig(t_symbol *s/*iemlib*/, int argc, t_atom *argv/*iemlib*/)
{
    t_voutlet *x = (t_voutlet *)pd_new(voutlet_class);
    /*iemlib*/
    int bcol= -262144;
    int lcol= -1;
    int a=IEM_GUI_DEFAULTSIZE;
    int ldx=8, ldy=-8;
    int f=0, fs=10;
    t_symbol *lab = gensym("empty");

    if((argc == 8)&&IS_A_FLOAT(argv,0)&&IS_A_SYMBOL(argv,1)
       &&IS_A_FLOAT(argv,2)&&IS_A_FLOAT(argv,3)&&IS_A_FLOAT(argv,4)
       &&IS_A_FLOAT(argv,5)&&IS_A_FLOAT(argv,6)&&IS_A_FLOAT(argv,7))
    {
        a = (int)atom_getintarg(0, argc, argv);
	lab = atom_getsymbolarg(1, argc, argv);
	ldx = (int)atom_getintarg(2, argc, argv);
	ldy = (int)atom_getintarg(3, argc, argv);
        f = (int)atom_getintarg(4, argc, argv);
        fs = (int)atom_getintarg(5, argc, argv);
        bcol = (int)atom_getintarg(6, argc, argv);
        lcol = (int)atom_getintarg(7, argc, argv);
    }
    x->x_glist = (t_glist *)canvas_getcurrent();
    x->x_lab = lab;
    x->x_ldx = ldx;
    x->x_ldy = ldy;
    if(f == 1) strcpy(x->x_font, "helvetica");
    else if(f == 2) strcpy(x->x_font, "times");
    else strcpy(x->x_font, "courier");
    if(fs < 4)
	fs = 4;
    x->x_fontsize = fs;
    x->x_size = iem_gui_clip_size(a&0xfffffffe);
    if(lcol < 0)
    {
        lcol = -1 - lcol;
        x->x_lcol = ((lcol & 0x3f000) << 6)|((lcol & 0xfc0) << 4)|((lcol & 0x3f) << 2);
    }
    else
    {
	lcol = iem_gui_modulo_color(lcol);
	x->x_lcol = iem_gui_color_hex[lcol];
    }
    if(bcol < 0)
    {
        bcol = -1 - bcol;
	x->x_bcol = ((bcol & 0x3f000) << 6)|((bcol & 0xfc0) << 4)|((bcol & 0x3f) << 2);
    }
    else
    {
	bcol = iem_gui_modulo_color(bcol);
	x->x_bcol = iem_gui_color_hex[bcol];
    }
    x->x_selected = 0;
    /*iemlib*/
    x->x_canvas = canvas_getcurrent();
    x->x_parentoutlet = canvas_addoutlet(x->x_canvas,
    	&x->x_obj.ob_pd, &s_signal);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
    x->x_endbuf = x->x_buf = (t_float *)getbytes(0);
    x->x_bufsize = 0;
    return (x);
}


static void voutlet_setup(void)
{
    voutlet_class = class_new(gensym("outlet"), (t_newmethod)voutlet_new,
    	(t_method)voutlet_free, sizeof(t_voutlet), CLASS_NOINLET, /*iemlib*/A_GIMME/*iemlib*/, 0);
    class_addcreator((t_newmethod)voutlet_newsig, gensym("outlet~"), 0);
    /*iemlib*/
    class_addcreator((t_newmethod)voutlet_new, gensym("ol"), A_GIMME, 0);
    class_addcreator((t_newmethod)voutlet_newsig, gensym("ol~"), A_GIMME, 0);
    /*iemlib*/
    class_addbang(voutlet_class, voutlet_bang);
    class_addpointer(voutlet_class, voutlet_pointer);
    class_addfloat(voutlet_class, (t_method)voutlet_float);
    class_addsymbol(voutlet_class, voutlet_symbol);
    class_addlist(voutlet_class, voutlet_list);
    class_addanything(voutlet_class, voutlet_anything);
    /*iemlib*/
    class_addmethod(voutlet_class, (t_method)voutlet_dialog, gensym("dialog"),
		    A_GIMME, 0);
    /*iemlib*/
    class_addmethod(voutlet_class, (t_method)voutlet_dsp, gensym("dsp"), 0);
    /*iemlib*/
    voutlet_widgetbehavior.w_getrectfn = voutlet_getrect;
    voutlet_widgetbehavior.w_displacefn = voutlet_displace;
    voutlet_widgetbehavior.w_selectfn = voutlet_select;
    voutlet_widgetbehavior.w_activatefn = NULL;
    voutlet_widgetbehavior.w_deletefn = voutlet_delete;
    voutlet_widgetbehavior.w_visfn = voutlet_vis;
#if (PD_VERSION_MINOR > 31)
    voutlet_widgetbehavior.w_clickfn = NULL;
    voutlet_widgetbehavior.w_propertiesfn = voutlet_properties;
#endif
    voutlet_widgetbehavior.w_savefn = voutlet_save;
    class_setwidget(voutlet_class, &voutlet_widgetbehavior);
    /*iemlib*/
    class_sethelpsymbol(voutlet_class, gensym("pd"));
}


/* ---------------------------- overall setup ----------------------------- */

void g_io_setup(void)
{
    vinlet_setup();
    voutlet_setup();
}
